#!/usr/bin/perl
# $Id$
print << '.';
/*
 * This header file is auto-generated from pttbbs/mbbsd/var.c .
 * Please do NOT edit this file directly.
 */

#ifndef INCLUDE_VAR_H
#define INCLUDE_VAR_H
#include "bbs.h"
.
while( <> ){
    if( /^\w/ ){
	chomp;
	s|//.*||;
	s|/\*.*?\*/||;
	s/\=\s*\{.*?\}//;

	s/\=.*?([,;])/$1/g;
	s/\=.*?\{//g;
	$_ .= ';' if( index($_, ';') == -1 );
	s/\=.*?([,;])/$1/g;
	s/\=.*?\{//g;
	print "extern $_\n";
    } elsif( /^\s*\#/ && !/include/ ){
	print;
    }
}
print "#endif /* INCLUDE_VAR_H */\n";

